from functools import partial
from typing import Callable, Dict, List, Optional

from . import models

WeightFunction = Callable[[Optional[models.Edge], models.Edge], float]

WEIGHT_FUNCTIONS: Dict[str, WeightFunction] = {}


def register(func: WeightFunction, function_name: str) -> None:
    """
    Register a custom weight function.
    """
    if function_name in WEIGHT_FUNCTIONS:
        raise ValueError("Function with the same name is already registered.")
    WEIGHT_FUNCTIONS[function_name] = func


def unregister(function_name: str) -> None:
    """
    Unregister a custom weight function.
    """
    try:
        del WEIGHT_FUNCTIONS[function_name]
    except KeyError:
        pass


def get_function(name: str) -> WeightFunction:
    """
    Return the matching weight function.
    """
    try:
        return WEIGHT_FUNCTIONS[name.lower()]
    except KeyError as error:
        raise ValueError("Weight function does not exist.") from error


def get_function_names() -> List[str]:
    """
    Return list of possible weight functions.
    """
    return list(WEIGHT_FUNCTIONS.keys())


def weight_by_attr(
    attr: str, prev_edge: Optional[models.Edge], edge: models.Edge
) -> float:
    """
    Generic weight function to retrieve a value from an edge.
    """
    return getattr(edge, attr)


travel_time = partial(weight_by_attr, "travel_time")
register(travel_time, "travel_time")

length = partial(weight_by_attr, "length")
register(length, "length")



from pydantic import BaseModel, validator

from .. import models, weights


class VersionResponse(BaseModel):
    version: str


class RouteRequest(BaseModel):
    origin: models.Location
    destination: models.Location
    weight: str

    @validator("weight")
    def validate_weight(cls, value: str) -> str:  # noqa: N805
        valid_values = weights.get_function_names()
        if value not in valid_values:
            raise ValueError(
                f"Invalid weight function. Possible values are: {valid_values}"
            )
        return value



import logging
import time
from typing import Any, Callable, TypeVar, cast

logger = logging.getLogger()

Func = TypeVar('Func', bound=Callable[..., Any])


def timeit(func: Func) -> Func:
    """
    Measure and log execution time of the decorated function.
    """

    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        logger.debug(f"{func.__qualname__}: {end - start} s")
        return result

    return cast(Func, wrapper)



import json
import logging
from pathlib import Path
from typing import Optional, Tuple

import click

from . import models
from .engine import Engine
from .utils import click as click_utils
from .utils import core as core_utils
from .utils import graph as graph_utils


def set_log_level(log_level: Optional[str]) -> None:
    if log_level:
        level = getattr(logging, log_level)
        logging.basicConfig(level=level)


@click.group()
def main() -> None:
    pass


@main.command()
@click.option('--log-level', type=click.Choice(["INFO", "DEBUG"]), default="INFO")
@click.option(
    '--api-key',
    type=str,
    default=None,
    help="Google API key to add elevation to nodes and grades to edges.",
)
@click.option(
    '-n', '--node-tags', multiple=True, type=str, help="Additional node tags to fetch."
)
@click.option(
    '-e', '--edge-tags', multiple=True, type=str, help="Additional edge tags to fetch."
)
@click.argument('regions', nargs=-1, type=str)
@click.argument('target', type=click_utils.Path(exists=False, dir_okay=False))
def download(
    regions: Tuple[str],
    target: Path,
    log_level: Optional[str],
    node_tags: Tuple[str],
    edge_tags: Tuple[str],
    api_key: Optional[str] = None,
) -> None:
    """
    Download a compatible map.
    """
    set_log_level(log_level)

    graph = graph_utils.download_map(
        list(regions),
        node_tags=list(node_tags),
        edge_tags=list(edge_tags),
        api_key=api_key,
    )
    graph_utils.save_map(graph, target)


@main.command()
@click.option('--log-level', type=click.Choice(["INFO", "DEBUG"]), default="INFO")
@click.argument('map_path', type=click_utils.Path(exists=True, dir_okay=False))
@click.argument('origin', type=click_utils.LocationParamType())
@click.argument('destination', type=click_utils.LocationParamType())
@click.argument('weight', type=str)
@click.argument('travel_time', type=str)
def route(
    map_path: Path,
    origin: models.Location,
    destination: models.Location,
    weight: str,
    travel_time: str,
    log_level: Optional[str],
) -> None:
    """
    Calculate a shortest path.

    \b
    MAP Path to an OSM graphml file. Format: .graphml
    ORIGIN GPS location. Format: latitude,longitude
    DESTINATION GPS location. Format: latitude,longitude
    WEIGHT Module path to weight function, eg. "routor.weights.length"
    TRAVEL_TIME Module path to weight function, which returns the travel time for an edge, eg. "routor.weights.travel_time"
    """
    set_log_level(log_level)

    # load weight functions
    weight_func = core_utils.import_weight_function(weight)
    travel_time_func = core_utils.import_weight_function(travel_time)

    # do routing
    engine = Engine(map_path)
    data = engine.route(origin, destination, weight_func, travel_time_func)

    print(json.dumps(data.dict(), indent=2))



import pathlib

import click

from .. import models


class LocationParamType(click.ParamType):
    name = "LATITUDE,LONGITUDE"

    def convert(self, value: str, param, ctx) -> models.Location:
        """
        Parse comma separated GPS coordinates.
        """
        try:
            latitude_str, longitude_str = value.split(',')
        except ValueError:
            self.fail("expected two items separated by comma", param, ctx)

        try:
            latitude = float(latitude_str)
            longitude = float(longitude_str)
        except ValueError:
            self.fail("Invalid number", param, ctx)

        return models.Location(latitude=latitude, longitude=longitude)


class Path(click.Path):
    """
    A click argument that returns a `pathlib.Path` instance.
    """

    def convert(self, value, param, ctx) -> pathlib.Path:  # type: ignore
        return pathlib.Path(super().convert(value, param, ctx))



import contextlib
import logging
from pathlib import Path
from typing import Any, Callable, Generator, List, Optional

import networkx
import osmnx

from .debug import timeit

logger = logging.getLogger()


@timeit
def load_map(map_path: Path) -> networkx.DiGraph:
    """
    Load graph from a .graphml file.
    """
    graph = osmnx.io.load_graphml(map_path)
    return graph


@contextlib.contextmanager
def osmnx_config(
    node_tags: List[str], edge_tags: List[str]
) -> Generator[None, None, None]:
    """
    Prepare osmnx config for downloading data.
    """
    # aggregate default tags
    useful_node_tags = set(
        osmnx.settings.useful_tags_node
        + osmnx.settings.osm_xml_node_attrs
        + osmnx.settings.osm_xml_node_tags
        + ["junction"]  # additional useful tags
        + ["street_count"]  # custom enhancements
        + node_tags
    )
    useful_edge_tags = set(
        osmnx.settings.useful_tags_way
        + osmnx.settings.osm_xml_way_attrs
        + osmnx.settings.osm_xml_way_tags
        + ["junction"]  # additional useful tags
        + ["bearing", "speed_kph", "travel_time"]  # custom enhancements
        + edge_tags
    )

    original_settings = {
        "all_oneway": osmnx.settings.all_oneway,
        "useful_tags_node": osmnx.settings.useful_tags_node,
        "useful_tags_way": osmnx.settings.useful_tags_node,
    }
    new_settings = {
        "all_oneway": False,  # we need digraph, an edge for each direction
        "useful_tags_node": useful_node_tags,
        "useful_tags_way": useful_edge_tags,
    }
    try:
        logger.debug("Update osmnx configuration", extra=new_settings)
        osmnx.config(**new_settings)
        yield
    finally:
        logger.debug("Resetting osmnx configuration", extra=original_settings)
        osmnx.config(**original_settings)


@timeit
def tag_roundabout_nodes(graph: networkx.DiGraph) -> None:
    """
    Tag nodes within a roundabout as junction=roundabout.

    Not every node within a roundabout is always tagged accordingly. Let's fix that.
    More information:
    * https://wiki.openstreetmap.org/wiki/Key:junction
    * https://wiki.openstreetmap.org/wiki/Tag:junction%3Droundabout
    """
    for u, v, edge_data in graph.edges(data=True):
        if edge_data.get("junction", None) != "roundabout":
            continue

        for node_id in (u, v):
            node_data = graph.nodes[node_id]
            if node_data.get("junction", None) == "circular":
                # no need to do anything as this is a "roundabout"
                # https://wiki.openstreetmap.org/wiki/Tag:junction%3Dcircular?
                continue

            if "junction" in node_data and node_data["junction"] != "roundabout":
                logger.warning(
                    "Node %s was already tagged with `junction='%s'`.",
                    node_id,
                    node_data["junction"],
                )
            node_data["junction"] = "roundabout"


@timeit
def download_map(
    location: List[str],
    node_tags: Optional[List[str]] = None,
    edge_tags: Optional[List[str]] = None,
    api_key: Optional[str] = None,
) -> networkx.DiGraph:
    """
    Download map from OSM for specific locations.
    """
    logger.info(f"Download map for {location}")
    return custom_graph_from_x(
        location, node_tags=node_tags, edge_tags=edge_tags, api_key=api_key
    )


def custom_graph_from_x(
    *args: List[Any],
    graph_from_x: Callable = osmnx.graph_from_place,
    node_tags: Optional[List[str]] = None,
    edge_tags: Optional[List[str]] = None,
    api_key: Optional[str] = None,
) -> networkx.DiGraph:
    """
    Download map from OSM.
    """
    with osmnx_config(node_tags or [], edge_tags or []):
        graph = graph_from_x(
            *args,
            network_type="drive",
            retain_all=False,  # only keep biggest connected network
            truncate_by_edge=True,  # Keep entirety of edges, rather than cropping at distance limit
            simplify=False,  # Do not correct and simplify street network topology
        )

    # add additional attributes
    logger.info("Enhance map with additional attributes")
    logger.info("> Tag node as roundabout")
    tag_roundabout_nodes(graph)

    logger.info("> Adding street count to nodes")
    street_count_data = osmnx.utils_graph.count_streets_per_node(graph)
    for node_id, street_count in street_count_data.items():
        graph.nodes[node_id]["street_count"] = street_count

    if api_key:
        logger.info("> Adding elevation")
        osmnx.add_node_elevations(graph, api_key, precision=5)

        logger.info("> Add edge grades")
        osmnx.elevation.add_edge_grades(graph)

    logger.info("> Adding bearing")
    osmnx.bearing.add_edge_bearings(graph)

    logger.info("> Adding edge speeds")
    osmnx.speed.add_edge_speeds(graph, hwy_speeds=None, fallback=30)

    logger.info("> Adding travel time")
    osmnx.speed.add_edge_travel_times(graph)

    return graph


@timeit
def save_map(graph: networkx.DiGraph, target: Path) -> None:
    """
    Save graph as .graphml file.
    """
    logger.info(f"Saving graph as {target.absolute()}.")
    osmnx.save_graphml(graph, filepath=str(target))



from routor.weights import WeightFunction


def import_weight_function(path: str) -> WeightFunction:
    """
    Imports a weight function from a string.
    """
    from importlib import import_module

    module_path, func_name = path.rsplit('.', 1)
    module = import_module(module_path)
    weight_func = getattr(module, func_name)

    return weight_func



from typing import Any, List, Optional

import networkx
from pydantic import BaseModel, Extra, Field, validator

from routor import exceptions


class Location(BaseModel):
    latitude: float  # alias: y
    longitude: float  # alias: x

    @validator("latitude")
    def validate_latitude(cls, value: float) -> float:  # noqa: N805
        if -90 > value or value > 90:
            raise ValueError("Latitude must be between -90 and 90")
        return value

    @validator("longitude")
    def validate_longitude(cls, value: float) -> float:  # noqa: N805
        if -180 > value or value > 180:
            raise ValueError("Longitude must be between -180 and 180")
        return value


class Node(Location):
    node_id: int
    osm_id: int = Field(alias="osmid")
    street_count: int

    elevation: Optional[float] = None  # only set if an API key was set during download

    class Config:
        allow_population_by_field_name = True
        extra = Extra.allow

    @classmethod
    def from_graph(cls, graph: networkx.Graph, node_id: int) -> "Node":
        try:
            node_data = graph.nodes[node_id]
        except KeyError as error:
            raise exceptions.NodeDoesNotExist() from error

        return cls(
            node_id=node_id,
            latitude=node_data["y"],
            longitude=node_data["x"],
            **node_data,
        )


class Edge(BaseModel):
    start: Node
    end: Node
    osm_id: int = Field(alias="osmid")
    oneway: bool
    length: float
    bearing: float
    travel_time: float
    geometry: Any  # TODO

    grade: Optional[float] = None  # only if elevation is present

    class Config:
        allow_population_by_field_name = True
        extra = Extra.allow

    @classmethod
    def from_graph(cls, graph: networkx.Graph, start_id: int, end_id: int) -> "Edge":
        start = Node.from_graph(graph, start_id)
        end = Node.from_graph(graph, end_id)
        return Edge.from_nodes(graph, start, end)

    @classmethod
    def from_nodes(cls, graph: networkx.DiGraph, start: Node, end: Node) -> "Edge":
        try:
            edge_data = graph[start.node_id][end.node_id][0]
        except KeyError as error:
            raise exceptions.EdgeDoesNotExist() from error

        return cls(start=start, end=end, **edge_data)


class Route(BaseModel):
    costs: float
    length: float
    travel_time: float
    path: List[Location]



import importlib
import logging
import pkgutil
from types import ModuleType
from typing import Dict

logger = logging.getLogger()


def discover_and_load_plugins() -> Dict[str, ModuleType]:
    """
    Discover and load routor plugins.
    """
    logger.debug("Loading plugins")

    plugins: Dict[str, ModuleType] = {}
    for _, name, _ in pkgutil.iter_modules():
        if not name.startswith('routor_'):
            continue

        module = importlib.import_module(name)
        plugins[name] = module
        logger.info("Loaded %s", name)

    return plugins


discovered_plugins = discover_and_load_plugins()



import logging
from pathlib import Path
from typing import Any, List, Optional, Tuple

import networkx
import osmnx
from more_itertools import pairwise
from networkx_astar_path import astar_path

from . import exceptions, models, weights
from .utils.debug import timeit
from .utils.graph import load_map

logger = logging.getLogger()


class Engine:
    graph: networkx.DiGraph

    @timeit
    def __init__(self, map_path: Path) -> None:
        logger.info("Initialise engine")
        self.graph = load_map(map_path)
        logger.info(
            f"Map loaded (edges: {len(self.graph.edges)}, nodes: {len(self.graph.nodes)})"
        )

    @timeit
    def find_path(
        self,
        origin: models.Node,
        destination: models.Node,
        weight: weights.WeightFunction,
    ) -> List[models.Node]:
        """
        Calculate a route using the given weight.
        """
        if origin.node_id not in self.graph.nodes:
            raise exceptions.NodeDoesNotExist(f"{origin} does not exists.")
        if origin.node_id not in self.graph.nodes:
            raise exceptions.NodeDoesNotExist(f"{origin} does not exists.")

        def _weight_wrapper(
            graph: networkx.DiGraph,
            prev_edge_nodes: Optional[Tuple[Any, Any]],
            edge_nodes: Tuple[Any, Any],
        ) -> float:
            prev_edge: Optional[models.Edge] = None
            if prev_edge_nodes:
                prev_edge = models.Edge.from_graph(self.graph, *prev_edge_nodes)
            edge = models.Edge.from_graph(self.graph, *edge_nodes)

            return weight(prev_edge, edge)

        logger.info(
            f"Calculating path from {origin.osm_id} to {destination.osm_id} with {weight}"
        )
        path = astar_path(
            self.graph, origin.node_id, destination.node_id, weight=_weight_wrapper
        )
        logger.info(f"Found path with {len(path)} items.")
        return [models.Node.from_graph(self.graph, node_id) for node_id in path]

    @timeit
    def route(
        self,
        origin: models.Location,
        destination: models.Location,
        weight_func: weights.WeightFunction,
        travel_time_func: weights.WeightFunction,
    ) -> models.Route:
        """
        Calculate a shortest path.
        """
        origin_node = self.get_closest_node(origin)
        destination_node = self.get_closest_node(destination)

        path = self.find_path(origin_node, destination_node, weight_func)
        costs = self.costs_for_path(path, weight_func)
        length = self.length_of_path(path)
        travel_time = self.travel_time_of_path(path, func=travel_time_func)

        route = models.Route(
            costs=round(costs, 2),
            length=round(length, 2),
            travel_time=round(travel_time, 2),
            path=[models.Location(**node.dict()) for node in path],
        )
        return route

    @timeit
    def costs_for_path(
        self, path: List[models.Node], func: weights.WeightFunction
    ) -> float:
        """
        Calculate the costs for a given path.
        """
        edges = (
            models.Edge.from_nodes(self.graph, start, end)
            for start, end in pairwise(path)
        )

        costs: float = 0
        for index, (prev_edge, edge) in enumerate(pairwise(edges)):
            if index == 0:
                costs = func(None, prev_edge)

            costs += func(prev_edge, edge)

        return costs

    @timeit
    def length_of_path(self, path: List[models.Node]) -> float:
        """
        Calculate the length of a given path.
        """
        edges = (
            models.Edge.from_nodes(self.graph, start, end)
            for start, end in pairwise(path)
        )

        return sum(edge.length for edge in edges)

    @timeit
    def travel_time_of_path(
        self, path: List[models.Node], func: weights.WeightFunction
    ) -> float:
        """
        Calculate the travel time of a given path.
        """
        return self.costs_for_path(path, func)

    @timeit
    def get_closest_node(self, location: models.Location) -> models.Node:
        """
        Get the closest node to a GPS location.
        """
        node_id = osmnx.get_nearest_node(
            self.graph, (location.latitude, location.longitude)
        )
        node = models.Node.from_graph(self.graph, node_id)
        logger.info(f"Found closest node for {location} is {node.osm_id}")
        return node



from pathlib import Path

from pydantic import BaseSettings, validator

from routor.utils import core as core_utils
from routor.weights import WeightFunction


class Settings(BaseSettings):
    map_path: Path
    travel_time_func: str = "routor.weights.travel_time"

    class Config:
        env_file = '.env'
        env_file_encoding = 'utf-8'

    @validator('travel_time_func')
    def _validate_travel_time_func(cls, func_path):  # noqa: N805
        """
        Make sure that the travel_function is importable
        """
        core_utils.import_weight_function(func_path)
        return func_path

    def get_travel_time_func(self) -> WeightFunction:
        """
        Get the configured travel time function.
        """
        return core_utils.import_weight_function(self.travel_time_func)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['routor', 'routor.api', 'routor.utils']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'fastapi>=0.63.0,<0.64.0',
 'more-itertools>=8.6.0,<9.0.0',
 'networkx-astar-path>=1.0.1,<2.0.0',
 'networkx>=2.5,<3.0',
 'osmnx>=0.16.2,<0.17.0',
 'pydantic>=1.7.3,<2.0.0',
 'python-dotenv>=0.15.0,<0.16.0',
 'uvicorn>=0.13.2,<0.14.0']

entry_points = \
{'console_scripts': ['routor = routor.cli:main']}

setup_kwargs = {
    'name': 'routor',
    'version': '0.5.1',
    'description': 'Simple osm routing engine.',
    'long_description': '# routor\n\n![PyPI](https://img.shields.io/pypi/v/routor?style=flat-square)\n![GitHub Workflow Status (main)](https://img.shields.io/github/workflow/status/escaped/routor/Test%20&%20Lint/main?style=flat-square)\n![Coveralls github branch](https://img.shields.io/coveralls/github/escaped/routor/main?style=flat-square)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/routor?style=flat-square)\n![PyPI - License](https://img.shields.io/pypi/l/routor?style=flat-square)\n\nSimple routing engine for OpenStreetMaps with easy to customize profiles/weight-functions. \n\n## Requirements\n\n* Python 3.6.1 or newer\n\n## Installation\n\n```sh\npip install routor\n```\n\n## Usage\n\n### CLI\n\nThe CLI offers multiple commands, use `routor --help` to find out more.\n\n#### Download map\n\nDownloads a compatible map from OSM, eg.\n\n```sh\nroutor download "Bristol, England" ./bristol.graphml\n```\n\nBy default it only adds a handful of tags ([nodes](https://github.com/gboeing/osmnx/blob/77b2535776b4397ae0deda402398609b3a4694a6/osmnx/settings.py#L5), [edge](https://github.com/gboeing/osmnx/blob/77b2535776b4397ae0deda402398609b3a4694a6/osmnx/settings.py#L49)) to the graph.\nUse `-n` or `-e` to add other available tags ([edge](https://github.com/gboeing/osmnx/blob/77b2535776b4397ae0deda402398609b3a4694a6/osmnx/settings.py#L29), [node](https://github.com/gboeing/osmnx/blob/77b2535776b4397ae0deda402398609b3a4694a6/osmnx/settings.py#L28)) as well.\nAdditionally, you can download multiple regions at once:\n\n```sh\nroutor download -n junction -n traffic_signals -e surface -e lanes "Bristol, England" "Somerset, England" ./bristol_somerset.graphml\n```\n\nBy default, each downloaded map is enhanced with\n\n* `street_count` - how many physical segments are connected to a node\n* `bearing` - angle of each edge\n* `speed_kph` - free-flow travel speed based on `maxspeed`, fallback is set to `30` kph (see [osmnx](https://osmnx.readthedocs.io/en/stable/osmnx.html#osmnx.speed.add_edge_speeds) for more information)\n* `travel_time` - Travel time based on `speed_kph` and `length`\n\nIf you provide a [Google API](https://developers.google.com/maps/documentation/javascript/get-api-key) (using --api-key), the following additional attributes are available:\n\n* `elevation` - elevation above sea level\n* `grade`/`grade_abs` - grade of an endge\n\n#### Calculate route\n\nDetermine the optimal route between two points using the given weight function and print the route as `JSON` to `stdout`.\n\n```sh\nroutor route -- ./bristol.graphml  "51.47967237816338,-2.6174926757812496" "51.45422084861252,-2.564105987548828" "routor.weights.length"\n```\n\n### Web API\n\n#### Configuration\n\nThe configuration is either read from a `.env` file or the environment.\nBefore you are able to run the server, you have to set the variables mentioned in [routor/api/config.py](routor/api/config.py).\n\n#### Run the API\n\nThe api is served using [uvicorn](https://www.uvicorn.org/).\nTo start the server run\n\n```sh\nuvicorn routor.api.main:app\n```\n\nThe API will be available at http://127.0.0.1:8000 and the docs at http://127.0.0.1:8000/docs.\n\n### As library\n\nYou can also use the engine as a library.\nTo calculate a route from A to B you can do\n\n```python\nfrom pathlib import Path\n\nfrom routor.engine import Engine\nfrom routor import models, weights\n\n...\nmap_path = Path(...)\nengine = Engine(map_path)\n\norigin = models.Location(latitude=51.47967237816338, longitude=-2.6174926757812496)\ndestination = models.Location(latitude=51.45422084861252, longitude=-2.564105987548828)\n\nroute = engine.route(origin, destination, weight_func=weights.length, travel_time_func=weights.travel_time)  # shortest distance\n```\n\n## Available weight-functions\n\n### `"length"` / `routor.weights.length`\n\nCalculates the shortest path from A to B, only the length of an edge is taken into account.\n\n### `"travel_time"` / `routor.weight.travel_time`\n\nCalculates the fastest route based on [travel time](https://osmnx.readthedocs.io/en/stable/osmnx.html#osmnx.speed.add_edge_travel_times).\n\n## Plugins\n\n`routor` implements a simple plugin mechanism.\nSimply create a new module with the prefix `routor_`, make it available (install it, `sys.path` hack or similar) and it will be automatically discovered and loaded.\nDepending on how you structure your module/plugin, you have to do the registration of the additional functionality in either `routor_YOUR_MODULE/__init__.py` or `routor_YOUR_MODULE.py`.\n\n### Register a new weight function\n\nExisting weight functions are defined in [routor/weights.py](routor/weights.py) and can be taken as reference.\nTo register a new function in your plugin, you have to implement something similar to\n\n```python\n# __init__.py\nfrom typing import Optional\n\nfrom routor.weights import register\nfrom routor import models\n\n\ndef my_weight_func(prev_edge: Optional[models.Edge], edge: models.Edge) -> float:\n    ...\n    return ...\n\n\nregister(my_weight_func, "weight_func")\n```\n\n## Development\n\nThis project uses [poetry](https://poetry.eustace.io/) for packaging and\nmanaging all dependencies and [pre-commit](https://pre-commit.com/) to run\n[flake8](http://flake8.pycqa.org/), [isort](https://pycqa.github.io/isort/),\n[mypy](http://mypy-lang.org/) and [black](https://github.com/python/black).\n\nAdditionally, [pdbpp](https://github.com/pdbpp/pdbpp) and [better-exceptions](https://github.com/qix-/better-exceptions) are installed to provide a better debugging experience.\nTo enable `better-exceptions` you have to run `export BETTER_EXCEPTIONS=1` in your current session/terminal.\n\nClone this repository and run\n\n```bash\npoetry install\npoetry run pre-commit install\n```\n\nto create a virtual enviroment containing all dependencies.\nAfterwards, You can run the test suite using\n\n```bash\npoetry run pytest\n```\n\nThis repository follows the [Conventional Commits](https://www.conventionalcommits.org/)\nstyle.\n\n### Cookiecutter template\n\nThis project was created using [cruft](https://github.com/cruft/cruft) and the\n[cookiecutter-pyproject](https://github.com/escaped/cookiecutter-pypackage) template.\nIn order to update this repository to the latest template version run\n\n```sh\ncruft update\n```\n\nin the root of this repository.\n',
    'author': 'Alexander Frenzel',
    'author_email': 'alex@relatedworks.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/escaped/routor',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0',
}


setup(**setup_kwargs)



import logging
from functools import lru_cache
from typing import List

from fastapi import Depends, FastAPI

from .. import engine
from .. import models as engine_models
from .. import weights
from . import config, models

logger = logging.getLogger()
app = FastAPI()


@lru_cache()
def get_settings() -> config.Settings:
    """
    Return application settings.
    """
    logger.debug("initialise settings")
    return config.Settings()


def get_engine(
    settings: config.Settings = Depends(get_settings),  # noqa: B008
) -> engine.Engine:
    """
    Return an initialised routing engine.

    This is a singletone and the engine is only initialised once.
    """
    cached_value = getattr(get_engine, "__cache", None)
    if not cached_value:
        logger.debug("initialise engine")
        cached_value = engine.Engine(settings.map_path)
        get_engine.__cache = cached_value  # type: ignore
    return cached_value


@app.get("/weights", response_model=List[str])
def read_weights() -> List[str]:
    """
    Return all possible weight functions.
    """
    return weights.get_function_names()


@app.get("/route", response_model=engine_models.Route)
def read_route(
    data: models.RouteRequest,
    engine: engine.Engine = Depends(get_engine),  # noqa: B008
    settings: config.Settings = Depends(get_settings),  # noqa: B008
) -> engine_models.Route:
    """
    Calculate a route from A to B.
    """
    weight_func = weights.get_function(data.weight)

    route = engine.route(
        data.origin, data.destination, weight_func, settings.get_travel_time_func()
    )
    return route
